import os
from pathlib import Path
from .consts import gitcmd as git
from .tools import _raise_error, safe_relative_to, is_empty_dir
from .repo import Repo
from contextlib import contextmanager
from .cachedir import _get_cache_dir
from .consts import REPO_TYPE_SUB
import click
from .tools import rmtree
from .tools import is_forced
from .tools import get_effective_state
from .tools import verbose


def _commit_submodule_inside_clean_but_not_linked_to_parent(main_repo, subrepo):
    """
    If the submodule is clean inside but is not committed to the parent
    repository, this module does that.
    """
    if subrepo.dirty:
        return False

    if not [
        x for x in main_repo.all_dirty_files if x.absolute() == subrepo.path.absolute()
    ]:
        return

    try:
        main_repo.X(*(git + ["add", subrepo.path]))
    except Exception:
        if os.getenv("GIMERA_FORCE") == "1":
            return
    sha = subrepo.hex
    main_repo.X(
        *(
            git
            + [
                "commit",
                "-m",
                (
                    f"gimera: updated submodule at {subrepo.path.relative_to(main_repo.path)} "
                    f"to latest version {sha}"
                ),
            ]
        )
    )


def _fetch_latest_commit_in_submodule(
    working_dir, main_repo, repo_yml, common_vars, update=False
):
    path = Path(working_dir) / repo_yml.path
    if not path.exists():
        return
    verbose(f"Fetching latest commit in submodule {path}")

    state = get_effective_state(main_repo.path, path, common_vars)
    parent_gimera = state["parent_gimera"]
    repo = Repo(state["parent_repo"])
    relpath = state["parent_repo_relpath"]
    subrepo = repo.get_submodule(relpath)
    if subrepo.dirty:
        if os.getenv("GIMERA_FORCE") != "1":
            _raise_error(
                f"Directory {repo_yml.path} contains modified "
                "files. Please commit or purge before!"
            )
    sha = repo_yml.sha if not update else None

    def _commit_submodule():
        _commit_submodule_inside_clean_but_not_linked_to_parent(repo, subrepo)
        if main_repo.path != repo.path:
            _commit_submodule_inside_clean_but_not_linked_to_parent(main_repo, repo)

    if sha:
        if not subrepo.contain_commit(sha):
            with _temporary_switch_remote_to_cachedir(main_repo, repo_yml, relpath):
                subrepo.X(*(git + ["fetch", "--all"]))

        if not subrepo.contain_commit(sha):
            _raise_error(
                f"SHA {sha} does not seem to belong to a "
                f"branch at module {repo_yml.path}"
            )

        subrepo.X(*(git + ["checkout", "-f", repo_yml.branch]))
        subrepo.out("git", "reset", "--hard", f"origin/{repo_yml.branch}").strip()
        sha_of_branch = subrepo.out("git", "rev-parse", repo_yml.branch).strip()
        if sha_of_branch == sha:
            subrepo.X(*(git + ["checkout", "-f", repo_yml.branch]))
        else:
            subrepo.X(*(git + ["checkout", "-f", sha]))
    else:
        try:
            subrepo.X(*(git + ["checkout", "-f", repo_yml.branch]))
            subrepo.X(*(git + ["pull"]))
        except Exception:  # pylint: disable=broad-except
            _raise_error(f"Failed to checkout {repo_yml.branch} in {path}")
        else:
            _commit_submodule()

    subrepo.X(*(git + ["submodule", "update", "--init", "--recursive"]))
    _commit_submodule()

    # check if sha collides with branch
    subrepo.X(*(git + ["clean", "-xdff"]))
    if not repo_yml.sha or update:
        subrepo.X(*(git + ["checkout", "-f", repo_yml.branch]))
        with _temporary_switch_remote_to_cachedir(repo, repo_yml, relpath):
            subrepo.pull(repo_yml=repo_yml)
        _commit_submodule()

    # update gimera.yml on demand
    repo_yml.sha = subrepo.hex


@contextmanager
def _temporary_switch_remote_to_cachedir(main_repo, repo_yml, relpath):
    with _get_cache_dir(main_repo, repo_yml) as cache_dir:
        main_repo.X(*(git + ["submodule", "set-url", relpath, f"file://{cache_dir}"]))
        try:
            yield
        finally:
            main_repo.X(*(git + ["submodule", "set-url", relpath, repo_yml.url]))


def _make_sure_subrepo_is_checked_out(working_dir, main_repo, repo_yml, common_vars):
    """
    Could be, that git submodule update was not called yet.
    """
    assert repo_yml.type == REPO_TYPE_SUB
    path = working_dir / repo_yml.path
    state = get_effective_state(main_repo.path, path, common_vars)
    if path.exists() and not is_empty_dir(path):
        return
    repo = Repo(state["parent_repo"])
    with _temporary_switch_remote_to_cachedir(
        repo, repo_yml, state["parent_repo_relpath"]
    ):
        repo.X(
            *(
                git
                + [
                    "submodule",
                    "update",
                    "--init",
                    "--recursive",
                    state["parent_repo_relpath"],
                ]
            )
        )

    if not path.exists():
        _raise_error(f"After submodule update the path {path} did not exist")


def _has_repo_latest_commit(repo, branch):
    out = repo.out(*(git + ["ls-remote", "origin", branch]))
    sha = out.splitlines()[0].strip().split()[0].strip()
    current = repo.out(*(git + ["rev-parse", branch])).splitlines()[0].strip()
    result = sha == current
    return result


def __add_submodule(root_dir, working_dir, repo, config, all_config, common_vars):
    if config.type != REPO_TYPE_SUB:
        return
    verbose(f"Adding submodule {config.path}")
    path = working_dir / config.path
    relpath = path.relative_to(repo.path)
    if path.exists():
        # if it is already a submodule, dont touch
        try:
            submodule = repo.get_submodule(relpath)
        except ValueError:
            repo.output_status()
            repo.please_no_staged_files()
            # remove current path

            dirty_files = [
                x
                for x in repo.all_dirty_files_absolute
                if safe_relative_to(x, repo.path / relpath)
            ]
            if dirty_files:
                if not is_forced():
                    _raise_error(
                        f"Dirty files exist in {repo.path / relpath}. Changes would be lost."
                    )

            if repo.lsfiles(relpath):
                repo.X(*(git + ["rm", "-f", "-r", relpath]))
            if (repo.path / relpath).exists():
                rmtree(repo.path / relpath)

            repo.clear_empty_subpaths(config)
            repo.output_status()
            if not [
                # x for x in repo.staged_files if safe_relative_to(x, repo.path / relpath)
                x
                for x in repo.all_dirty_files
                if safe_relative_to(x, repo.path / relpath)
            ]:
                if relpath.exists():
                    # in case of deletion it does not exist
                    repo.X(*(git + ["add", relpath]))
            if repo.staged_files:
                repo.X(
                    *(
                        git
                        + [
                            "commit",
                            "-m",
                            f"removed path {relpath} to insert submodule",
                        ]
                    )
                )

            # make sure does not exist; some leftovers sometimes
            repo.force_remove_submodule(relpath)

        else:
            # if submodule points to another url, also remove
            if submodule.get_url(noerror=True) != config.url:
                repo.force_remove_submodule(submodule.path.relative_to(repo.path))
            else:
                return
    else:
        repo.force_remove_submodule(relpath)

    repo._fix_to_remove_subdirectories(all_config)
    if (repo.path / relpath).exists():
        # helped in a in the wild repo, where a submodule was hidden below
        repo.X(*(git + ["rm", "-rf", relpath]))
        rmtree(repo.path / relpath)

    with _get_cache_dir(repo, config) as cache_dir:
        repo.submodule_add(config.branch, str(cache_dir), relpath)
        repo.X(*(git + ["submodule", "set-url", relpath, config.url]))
        repo.X(*(git + ["add", ".gitmodules"]))
        click.secho(f"Added submodule {relpath} pointing to {config.url}", fg="yellow")
        if repo.staged_files:
            repo.X(*(git + ["commit", "-m", f"gimera added submodule: {relpath}"]))

    # check for success
    state = get_effective_state(
        root_dir,
        path,
        common_vars,
    )
    assert state["is_submodule"]



import subprocess
from pathlib import Path
from .consts import gitcmd as git
from .repo import Repo
from .tools import get_nearest_repo
from .tools import safe_relative_to
import os
import uuid
from datetime import datetime
import shutil
from .patches import remove_file_from_patch

to_cleanup = []


def get_snapshots(root_dir):
    path = root_dir / ".gimera" / "snapshots"
    res = []
    for snapshot in reversed(list(path.glob("*"))):
        res.append(snapshot.name)
    return res


def _get_repo_for_filter_paths(root_dir, filter_paths):
    repos = set()
    for path in filter_paths:
        repo = get_nearest_repo(root_dir, path)
        repos.add(repo)
    if len(repos) > 1:
        raise NotImplementedError("Two many repos associated.")
    repo = list(repos)[0]
    return repo


def snapshot_recursive(root_dir, filter_paths, token=None):
    assert isinstance(filter_paths, list)

    _snapshot_dir(
        root_dir,
        path=None,
        filter_paths=filter_paths,
    )
    return _get_token()


def snapshot_restore(root_dir, filter_paths, token=None):
    token = token or _get_token()
    patches = root_dir / ".gimera" / "snapshots" / token

    for patchfile in patches.rglob("**/*.patch"):
        relpath = safe_relative_to(patchfile.parent, root_dir)
        relpath = Path("/".join(relpath.parts[3:])) / patchfile.stem

        assert all(
            str(x).startswith("/") for x in filter_paths
        ), "Only absolute paths please"

        if filter_paths and not [
            (str(root_dir / relpath / patchfile.stem)).startswith(str(x))
            for x in filter_paths
        ]:
            continue

        repo_dir = root_dir / relpath
        # find out the belonging repository; if switched from submodule to integrated
        # a new root must be added to the patch
        nearest_repo_path = get_nearest_repo(root_dir, repo_dir)
        delta_path = safe_relative_to(repo_dir, nearest_repo_path)
        cmd = git + ["apply", "--reject"]
        if str(delta_path) != ".":
            cmd += ["--directory", delta_path]
        cmd += [patchfile]
        subprocess.check_call(cmd, cwd=nearest_repo_path)


def _find_matching_dirs(root_dir, path, filter_paths, cache=None):
    cache = cache or {}
    if path is None:
        path = root_dir

    cache.setdefault("nearest", {})
    repo = get_nearest_repo(root_dir, path)

    def _matches_filter_paths(path, direction):
        if direction == "before":
            return bool(any(x for x in filter_paths if safe_relative_to(x, path)))
        else:
            return bool(any(x for x in filter_paths if safe_relative_to(path, x)))

    before = _matches_filter_paths(path, "before")
    after = _matches_filter_paths(path, "after")

    if before or after:
        cache.setdefault("dirty", {})
        if repo not in cache["dirty"]:
            cache["dirty"][repo] = Repo(repo).all_dirty_files
        if cache["dirty"][repo]:
            yield Repo(repo), path
        for sub in path.iterdir():
            if sub.is_dir():
                if sub.name == ".git":
                    continue
                yield from _find_matching_dirs(root_dir, sub, filter_paths, cache=cache)


def _snapshot_dir(root_dir, path, filter_paths=None):
    matching_dirs = list(
        reversed(list(_find_matching_dirs(root_dir, path, filter_paths)))
    )

    cache = {"dirty": {}}
    for repo, path in matching_dirs:
        if repo.path not in cache["dirty"]:
            repo.X("git", "reset", output=True)
            cache["dirty"][repo.path] = repo.all_dirty_files_absolute

        dirty_files = cache["dirty"][repo.path]
        if not dirty_files:
            continue
        dirty_files = [x for x in dirty_files if x.parent == path]
        if not dirty_files:
            continue
        cache["dirty"].pop(repo.path)
        for dirty_file in dirty_files:
            if any(x == dirty_file.name for x in [".gitmodules", ".git"]):
                continue
            subprocess.check_call((git + ["add", dirty_file]), cwd=path)
            del dirty_file
        patch_file_content = subprocess.check_output(
            (git + ["diff", "--cached", "--relative"]), cwd=path
        )
        if patch_file_content:
            cache_file = _get_patch_filepath(root_dir, path)
            cache_file.write_bytes(patch_file_content)

    repos = map(Repo, set(x[0].path for x in matching_dirs))
    for repo in repos:
        subprocess.check_call((git + ["reset"]), cwd=repo.path)
        subprocess.check_call((git + ["checkout", "."]), cwd=repo.path)
        for dirtyfile in repo.untracked_files_absolute:
            if dirtyfile.is_dir():
                shutil.rmtree(dirtyfile)
            else:
                dirtyfile.unlink()


def cleanup():
    for path in to_cleanup:
        if path.exists():
            if path.is_file():
                path.unlink()
            else:
                shutil.rmtree(path)


def _get_token():
    if not os.getenv("GIMERA_TOKEN"):
        os.environ["GIMERA_TOKEN"] = (
            datetime.now().strftime("%Y%m%d-%H%M%S") + "-" + str(uuid.uuid4())
        )
    return os.getenv("GIMERA_TOKEN")


def _get_patch_filepath(root_dir, file_relpath):
    token = _get_token()
    file_relpath = safe_relative_to(file_relpath, root_dir)
    path = root_dir / '.gimera' / "snapshots" / token / f"{file_relpath}.patch"
    to_cleanup.append(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    return path


def list_snapshots(root_dir):
    res = []
    for dir in (root_dir / ".gimera" / "snapshots").glob("*"):
        if not dir.is_dir():
            continue
        parts = dir.parts
        parts = parts[parts.index("snapshots") + 1 :]
        res.append("/".join(parts))
    return res



from contextlib import contextmanager
import os
import click
from pathlib import Path
from .repo import Repo, Remote
from .tools import _raise_error
from .consts import gitcmd as git
from .tools import wait_git_lock
from .tools import rmtree
from .consts import REPO_TYPE_INT, REPO_TYPE_SUB
from .patches import _apply_patches
from .patches import _apply_patchfile
from .tools import _get_remotes
from .tools import get_effective_state
from .tools import get_nearest_repo
from .tools import verbose
from .patches import _apply_patchfile
from .cachedir import _get_cache_dir


def _update_integrated_module(
    working_dir,
    main_repo,
    repo_yml,
    update,
    common_vars,
    **options,
):
    """
    Put contents of a git repository inside the main repository.
    """
    # use a cache directory for pulling the repository and updating it
    with _get_cache_dir(main_repo, repo_yml, update=update) as cache_dir:
        if not os.access(cache_dir, os.W_OK):
            _raise_error(f"No R/W rights on {cache_dir}")
        repo = Repo(cache_dir)
        verbose(f"Updating integrated module {repo_yml.path}")

        parent_repo = main_repo
        dest_path = Path(working_dir) / repo_yml.path
        parent_repo = Repo(get_nearest_repo(main_repo.path, dest_path))

        # BTW: delete-after cannot remove unused directories - cool to know; is
        # just standarded out
        if dest_path.exists():
            rmtree(dest_path)

        with wait_git_lock(cache_dir):
            commit = repo_yml.sha or repo_yml.branch if not update else repo_yml.branch
            with repo.worktree(commit) as worktree:
                new_sha = worktree.hex
                msgs = [f"Updating submodule {repo_yml.path}"] + _apply_merges(
                    worktree, repo_yml
                )
                worktree.move_worktree_content(dest_path)
                # TODO perhaps not necessary as of line 63 -- seems to be necessary
                # case: submodule is in .gitignore; updates the submodule
                # then git add <path> needs to add the deleted files
                # Could also be that a subgimera sha was updated
                parent_repo.commit_dir_if_dirty(dest_path, "\n".join(msgs), force=True)
            del repo

        # apply patches:
        if os.getenv("GIMERA_DO_NOT_APPLY_PATCHES") != "1":
            _apply_patches(repo_yml)
        msg = f"updated {REPO_TYPE_INT} submodule: {repo_yml.path}"
        repo_yml.sha = new_sha
        if repo_yml.config.config_file in parent_repo.all_dirty_files_absolute:
            # could be, that the parent path of the gimera.yml belongs to gitignore
            # so force add
            parent_repo.X(*(git + ["add", '-f', repo_yml.config.config_file]))
        parent_repo.commit_dir_if_dirty(dest_path, msg)
        if any(
            str(x).startswith(str(dest_path)) for x in parent_repo.all_dirty_files_absolute
        ):
            parent_repo.X(*(git + ["add", dest_path]))

        if parent_repo.staged_files:
            parent_repo.X(*(git + ["commit", "-m", msg]))

        if repo_yml.edit_patchfile:
            _apply_patchfile(
                repo_yml.edit_patchfile_full_path, repo_yml.fullpath, error_ok=True
            )


def _apply_merges(repo, repo_yml):
    if not repo_yml.merges:
        return []

    configured_remotes = list(_get_remotes(repo_yml))
    # as we clone into a temp directory to allow parallel actions
    # we set the origin to the repo source
    configured_remotes.append(Remote(repo, "origin", repo_yml.url))
    for remote in configured_remotes:
        if list(filter(lambda x: x.name == remote.name, repo.remotes)):
            repo.set_remote_url(remote.name, remote.url)
        repo.add_remote(remote, exist_ok=True)

    remotes = []
    msg = []
    for remote, ref in repo_yml.merges:
        msg.append(f"Merging {remote} {ref}")
        click.secho(msg[-1])
        remote = [x for x in reversed(configured_remotes) if x.name == remote][0]
        repo.pull(remote=remote, ref=ref)
        remotes.append((remote, ref))
    return msg



import os
import time
import errno
 
class FileLockException(Exception):
    pass
 
class FileLock(object):
    """ A file locking mechanism that has context-manager support so 
        you can use it in a with statement. This should be relatively cross
        compatible as it doesn't rely on msvcrt or fcntl for the locking.
    """
 
    def __init__(self, file_name, timeout=10, delay=.05):
        """ Prepare the file locker. Specify the file to lock and optionally
            the maximum timeout and the delay between each attempt to lock.
        """
        if timeout is not None and delay is None:
            raise ValueError("If timeout is not None, then delay must not be None.")
        self.is_locked = False
        self.lockfile = os.path.join(os.getcwd(), "%s.lock" % file_name)
        self.file_name = file_name
        self.timeout = timeout
        self.delay = delay
 
 
    def acquire(self):
        """ Acquire the lock, if possible. If the lock is in use, it check again
            every `wait` seconds. It does this until it either gets the lock or
            exceeds `timeout` number of seconds, in which case it throws 
            an exception.
        """
        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.lockfile, os.O_CREAT|os.O_EXCL|os.O_RDWR)
                self.is_locked = True #moved to ensure tag only when locked
                break;
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if self.timeout is None:
                    raise FileLockException("Could not acquire lock on {}".format(self.file_name))
                if (time.time() - start_time) >= self.timeout:
                    raise FileLockException("Timeout occured.")
                time.sleep(self.delay)
#        self.is_locked = True
 
 
    def release(self):
        """ Get rid of the lock by deleting the lockfile. 
            When working in a `with` statement, this gets automatically 
            called at the end.
        """
        if self.is_locked:
            os.close(self.fd)
            os.unlink(self.lockfile)
            self.is_locked = False
 
 
    def __enter__(self):
        """ Activated when used in the with statement. 
            Should automatically acquire a lock to be used in the with block.
        """
        if not self.is_locked:
            self.acquire()
        return self
 
 
    def __exit__(self, type, value, traceback):
        """ Activated at the end of the with statement.
            It automatically releases the lock if it isn't locked.
        """
        if self.is_locked:
            self.release()
 
 
    def __del__(self):
        """ Make sure that the FileLock instance doesn't leave a lockfile
            lying around.
        """
        self.release()


import os
import uuid
import click
import json
import shutil
import sys
import subprocess
from pathlib import Path
from .consts import gitcmd as git
from .repo import Repo
from .tools import prepare_dir
from .tools import remember_cwd
from .tools import reformat_url
from .tools import _raise_error
from .tools import rmtree
from .tools import replace_dir_with

# store big repos in tar file and try to restore from there;
# otherwise lot of downloads have to be done


from contextlib import contextmanager


@contextmanager
def _get_cache_dir(main_repo, repo_yml, no_action_if_not_exist=False, update=None):
    url = repo_yml.url
    if not url:
        click.secho(f"Missing url: {json.dumps(repo_yml, indent=4)}")
        sys.exit(-1)

    try:
        urlsafe = reformat_url(url, "git")
    except:
        urlsafe = url

    for c in "?:+[]{}\\/\"'_":
        urlsafe = urlsafe.replace(c, "-")
    urlsafe = urlsafe.split("@")[-1]

    path = Path(os.path.expanduser("~/.cache/gimera")) / urlsafe
    golden_path = path
    if no_action_if_not_exist and not golden_path.exists():
        yield None
        return
    possible_temp_path = Path(str(path) + "." + str(uuid.uuid4()))
    del path
    try:
        golden_path.parent.mkdir(exist_ok=True, parents=True)

        must_exist = ["HEAD", "refs", "objects", "config", "info"]
        if golden_path.exists() and any(
            not (golden_path / x).exists() for x in must_exist
        ):
            rmtree(golden_path)

        just_cloned = False
        if not golden_path.exists():
            click.secho(
                f"Caching the repository {repo_yml.url} for quicker reuse",
                fg="yellow",
            )
            tarfile = _get_cache_dir_tarfile(golden_path)
            with prepare_dir(possible_temp_path) as _path:
                with remember_cwd(
                    "/tmp"
                ):  # called from other situations where path may not exist anymore
                    if tarfile.exists():
                        _extract_tar_file(_path, tarfile)
                        just_cloned = True
                    else:
                        Repo(main_repo.path).X(*(git + ["clone", "--bare", url, _path]))
                        _make_tar_file(_path, tarfile)
                        just_cloned = True

        effective_path = possible_temp_path if just_cloned else golden_path

        if repo_yml.sha:
            repo = Repo(effective_path)
            if not repo.contain_commit(repo_yml.sha):
                # make a fetch quickly; sha is missing
                repo.X(*(git + ["fetch", "--all"]))
                if not repo.contain_commit(repo_yml.sha):
                    if not update:
                        _raise_error(
                            (
                                f"After fetching the commit {repo_yml.sha} "
                                f"was not found for {repo_yml.path}"
                            )
                        )

        yield effective_path

        if just_cloned:
            replace_dir_with(possible_temp_path, golden_path)

    finally:
        possible_temp_path = Path(possible_temp_path)
        if possible_temp_path.exists():
            rmtree(possible_temp_path)


def _get_cache_dir_tarfile(_path):
    return Path(str(_path) + ".tar.gz")


def _make_tar_file(_path, tarfile):
    if tarfile.exists():
        tarfile.unlink()
    click.secho(
        f"Creating tar file {tarfile} from {_path} - might take some time on large repos.",
        fg="yellow",
    )
    subprocess.check_call(["tar", "cfz", str(tarfile), "-C", str(_path), "."])


def _extract_tar_file(_path, tarfile):
    click.secho(f"Extracting tar file {tarfile} to {_path}", fg="yellow")
    subprocess.check_call(["tar", "xfz", str(tarfile)], cwd=_path)
